"""
Random utils.
"""

import random as core_random
import numpy.random as np_random

core_random = np_random

seed = core_random.seed
shuffle = core_random.shuffle
choices = core_random.choice
random = core_random.random

get_state = core_random.get_state
set_state = core_random.set_state



'''
Support classes for FASTENER.
'''

import itertools
from dataclasses import dataclass, field
from typing import List, Optional, Callable, Dict, Tuple, Any, Set
from abc import ABC, abstractmethod
import sklearn.feature_selection
import numpy as np

import fastener.random_utils as random_utils

Genes = List[bool]
FitnessFunction = Callable[["Genes"], Tuple["Result", Any]]
#for each item size it keeps lists of Items of this size
Population = Dict[int, List["EvalItem"]]
UnevaluatedPopulation = Dict[int, List["Item"]]


IntScaling = Callable[[int], int]
FloatScaling = Callable[[float], float]


def flatten_population(population: Population) -> List["EvalItem"]:
    '''Creates array form dictionary.

    Transforms dictionary with List values into a single List with
    values form the other lists.

    Args:
        population: Dictionary to be transformed.

    Returns:
        A List with values from all the other lists.
    '''
    return list(itertools.chain(*population.values()))


@dataclass
class Item:
    '''Class representing individual.

    An object that represents an individual in genethic algorithm
    (a set of features to be used) and can be evaluated.

    Attributes:
        genes: An array of booleans marking which features are are
            selected.
        size: An integer count of selected features
        number: An integer which written in binary represents the
            active features (in reversed order).
        generation: An integer representing the generation the
            individual was created in.
        parent_a: An array of booleans representing genes from a parent
            (if parent exists).
        parent_b: An array of booleans representing genes from a parent
            (if parent exists).

    '''

    genes: Genes
    size: int = field(init=False)
    # Number representing genes
    # Number is in little endian encoding, so that increasing the number of
    # genes (padding at the end) does not change number (appending new features
    # to existing ones keeps current caching valid)
    number: int = field(init=False)
    generation: int

    # parent_a: Optional["Item"]
    # parent_b: Optional["Item"]
    parent_a: Optional[Genes]
    parent_b: Optional[Genes]

    def __post_init__(self) -> None:
        '''Post init function.

        After genes, geenration and parents are set size and number
        are calculated
        '''
        self.size = sum(self.genes)
        self.number = Item.to_number(self.genes)

    def evaluate(self, fitness_function: FitnessFunction) -> "EvalItem":
        '''Evaluates item.

        Calculates the result of the feature set and returns an
        EvalItem object (Item with result attribute).

        Args:
            fitness_function: Function used to calculate result.

        Returns:
            An EvalItem object.
        '''
        result = fitness_function(self.genes)[0]
        return EvalItem(self.genes, self.generation, self.parent_a,
                        self.parent_b, result)

    def __eq__(self, other: Any) -> bool:
        '''Checks if this item object is equivalent to other.

        Checks if passed variable (other) is of type Item and if it is
        equivalent to the self object.

        Args:
            other: A variable for which we wish to check if it is
                equivalent to self.

        Returns:
            True if the passed variable is of type Item and is
            equivalent to self and False otherwise.
        '''
        return isinstance(other, Item) and self.number == other.number

    def __hash__(self) -> int:
        '''Returns hash value of number attribute.
        '''
        return hash(self.number)

    @classmethod
    def num_to_genes(cls, num: int, num_genes: int) -> "Genes":
        '''Converts number to genes array it represents.

        Converts number to binary and then sets the feature values
        accordingly to the string of ones and zeros. At the end it adds
        zeros if needed (for the not selected features)

        Args:
            num: The number we wish to convert to genes array.
            num_genes: Number of all genes (active and inactive).

        Returns:
            An array of booleans (genes), for the given number.
        '''
        return list(
            map(lambda x: bool(int(x)),
                reversed(bin(num)[2:].zfill(num_genes))))

    @classmethod
    def to_number(cls, genes: Genes) -> int:
        '''Converts array of genes to number.

        Loops throug the array of booleans and builds a string of ones
        and zeros then converts it to a decimal.

        Args:
            genes: An array of booleans (genes) to be converted to
                number.

        Returns:
            An integer representing the selected features.
        '''
        return int(''.join(map(str, map(int, reversed(genes)))), 2)

    @staticmethod
    def gene_union(g1: Genes, g2: Genes) -> Genes:
        '''Calculates a union of g1 and g2.

        Creates a new array of booleans in which i-th gene is True
        if i-th gene in g1 is True or i-th gene in g2 is True. It is
        used as a mating strategy.

        Args:
            g1: An array of booleans (genes).
            g2: An array of booleans (genes).

        Returns:
            An array of booleans (genes) in which i-th gene is True
            if i-th gene in g1 is True or i-th gene in g2 is True.
        '''
        return [i or j for i, j in zip(g1, g2)]

    @staticmethod
    def gene_intersection(g1: Genes, g2: Genes) -> Genes:
        '''Calculates an intersection of g1 and g2.

        Creates a new array of booleans in which i-th gene is True
        if i-th gene in g1 is True and i-th gene in g2 is True. It is
        used as a mating strategy.

        Args:
            g1: An array of booleans (genes).
            g2: An array of booleans (genes).

        Returns:
            An array of booleans (genes) in which i-th gene is True
            if i-th gene in g1 is True and i-th gene in g2 is True.
        '''
        return [i and j for i, j in zip(g1, g2)]

    @classmethod
    def from_genes(cls, genes: Genes, generation: int = 0) -> "Item":
        '''Creates Item from genes.

        Creates Item object with specified gened and generation.

        Args:
            genes: An boolean array of "genes".
            generation: An integer representing the generation we want
                the created item to be.

        Returns:
            An Item object we created from genes.
        '''
        return Item(genes, generation, None, None)

    @classmethod
    def indexed_symmetric_difference(cls, g1: Genes, g2: Genes) -> List[int]:
        '''Calculates indexes where genes differ.

        Loops through a pair of genes and stores indexes of places
        where the two differ in an array.

        Args:
            g1: An array of booleans (genes).
            g2: An array of booleans (genes).

        Returns:
            An integer array of indexes where the boolean array differ
        '''
        # XOR
        return [ind for ind, (i, j) in enumerate(zip(g1, g2)) if (i != j)]


@dataclass
class EvalItem(Item):
    '''Evaluated item.

    A child class of Item object (has all of the attributes + result).

    Attributes:
        result: A Result object containing a score of the evaluatin
            function for this set of attributes.

    '''

    result: "Result"

    # Same sized items are compared according to their result
    def __lt__(self, other: "EvalItem") -> bool:
        '''Check if other EvalItem object has larger score.

        Checks if self object has the same number of attributes
        (as other) and the score of self object is lower.

        Args:
            other: An EvalItem object to be compared.

        Returns:
            True if self and other object have the same size and self
            result is lower than others.

        Raises:
            AssertionError: If objects have different size attributes
        '''
        assert self.size == other.size
        return self.size >= other.size and self.result < other.result

    def __eq__(self, other: Any) -> bool:
        '''Overrides __eq__ methode from Item class.

        Checks if passed variable (other) is of type EvalItem and if it is
        equivalent to the self object.

        Args:
            other: A variable for which we wish to check if it is
                equivalent to self.

        Returns:
            True if the passed variable is of type EvalItem and is
            equivalent to self and False otherwise.
        '''
        return isinstance(other, EvalItem) and self.number == other.number

    def __hash__(self) -> int:
        '''Returns hash value of number attribute.
        '''
        return hash(self.number)

    def evaluate(self, fitness_function: FitnessFunction) -> "EvalItem":
        '''Overrides Evaluate method from Item class.

        Calculates the result of the feature set and returns an
        EvalItem object (Item with result attribute).

        Args:
            fitness_function: Function used to calculate result.

        Returns:
            Object self.
        '''
        return self

    def pareto_better(self, other: "EvalItem") -> bool:
        """Compares EvalItem objects.

        Checks if self has less (or equal) genes, but better
        (or at least similar result).

        Attributes:
            other: EvalItem object to be compared.

        Returns:
            True if self objects has less (or equal) features but
            larger (or equal) score and False otherwise.
        """
        return self.size <= other.size and other.result <= self.result


@dataclass
class Result:
    '''Result of an evaluated Item.

    Attributes:
        score: A float that holds score of evaluated function for a
            given set of features.
    '''

    score: float

    def __lt__(self, other: "Result") -> bool:
        return self.score < other.score

    def __le__(self, other: "Result") -> bool:
        return self.score <= other.score


class MatingStrategy(ABC):
    '''Mating strategy.

    An abstract class that defines the base of mate function but not
    how the mating is implemented.
    '''

    @abstractmethod
    def mate_internal(self, item1: Item, item2: Item) -> Genes:
        pass

    def mate(self, item1: Item, item2: Item, generation: int) -> Item:
        '''Initailizes mating and creates a new Item.

        Initializes mating strategy (subclass specific) with parents
        and from the output genes creates a new Item.

        Attributes:
            item1: First parent (Item object) to be used for mating.
            item2: Second parent (Item object) to be used for mating.
            generatinon: Integer representing generation of the
                offspring (result of mating).

        Returns:
            Item object that is the result of mating.
        '''
        genes = self.mate_internal(item1, item2)

        item = Item(genes, generation, item1.genes, item2.genes)

        return item

    def use_data_information(self, train_data: np.array,
                             train_target: np.array) -> None:
        pass


class UnionMating(MatingStrategy):
    '''Union mating strategy.

    A subclass of MatingStrategy which defines mating strategy as a
    union between genes (if at least one of parents has the gene, the
    decandant will too).
    '''
    def mate_internal(self, item1: Item, item2: Item) -> Genes:
        '''Implementation of union mating.

        Calls a static function of Item object which creates an array
        of genes with or logic and returns it.

        Args:
            item1: An Item object we wish to mate.
            item2: An Item object we wish to mate.

        Returns:
            An array of booleans (genes) that is the result of mating.
        '''
        return Item.gene_union(item1.genes, item2.genes)


class IntersectionMating(MatingStrategy):
    '''Intersection mating strategy.

    A subclass of MatingStrategy which defines mating strategy as an
    intersection between genes (only if both parents have the gene, the
    decandant will too).
    '''
    def mate_internal(self, item1: Item, item2: Item) -> Genes:
        '''Implementation of intersection mating.

        Calls a static function of Item object which creates an array
        of genes with and logic and returns it.

        Args:
            item1: An Item object we wish to mate.
            item2: An Item object we wish to mate.

        Returns:
            An array of booleans (genes) that is the result of mating.
        '''
        return Item.gene_intersection(item1.genes, item2.genes)


class IntersectionMatingWithInformationGain(IntersectionMating):
    '''Intersection mating strategy with information gain.

    A subclass of IntersectionMating. It chooses active features with
    intersection plus some features from either one of the parents,
    that have the highest information gain.  
    '''
    def __init__(self, number: Optional[IntScaling] = None, regression: bool = False) -> None:
        '''Init function.

        Inits parameters, if model uses regression it sets regression
        information gain, otherwise classification.

        Args:
            number: A function that decides how many features are to be
                added to the intersection.
            regression: A boolean telling us if model is ussing
                regression
        
        Returns:
        '''
        self.number = number or self.default_number
        self.scikit_information_gain: List[float] = []
        self.informationGainAlgorithm = sklearn.feature_selection.mutual_info_regression if regression else sklearn.feature_selection.mutual_info_classif

    @staticmethod
    def default_number(x: int) -> int:
        return min(int(x / 2) + 1, x)

    def use_data_information(self, train_data: np.array,
                             train_target: np.array) -> None:
        '''Computes information gain with train data.

        Computes information gain with train data, using the correct
        algorithm (depending on whether the model uses classifocation
        or regression).

        Args:
            train_data: Data with which the information gain is
                calculated.
            train_target: Target with which the information gain is
                calculated.
        
        Returns:
        '''
        self.scikit_information_gain = self.informationGainAlgorithm(train_data, train_target)

    def mate_internal(self, item1: Item, item2: Item) -> Genes:
        '''Implemented intersection mating with information gain.

        First Calculates intersection of genes. Then featuresare sorted
        by information gain and best (how many is decided by number
        function) features by information gain are added to the set.

        Args:
            item1: An Item object we wish to mate.
            item2: An Item object we wish to mate.

        Returns:
            An array of booleans (genes) that is the result of mating.
        '''
        genes = super().mate_internal(item1, item2)
        ma = max(item1.size, item2.size)
        mi = min(item1.size, item2.size)

        sy_difference = Item. \
            indexed_symmetric_difference(item1.genes, item2.genes)

        sy_difference.sort(key=lambda x: self.scikit_information_gain[x],
                           reverse=True)
        for j in range(self.number(ma - mi)):
            genes[sy_difference[j]] = True
        return genes


class IntersectionMatingWithWeightedRandomInformationGain(
    IntersectionMatingWithInformationGain):
    '''Intersection mating strategy with weighted information gain.

    A subclass of IntersectionMatingWithInformationGain. It chooses
    active features with intersection plus it selects some features
    (from either one of the parents) with the probability, 
    proportionate to the feature's information gain.    
    '''
    def __init__(self, number: Optional[IntScaling] = None,
                 scaling=None, regression: bool = False) -> None:
        super().__init__(number, regression)
        self.number = number or self.default_number
        self.scaling = scaling or self.default_scaling

    @staticmethod
    def default_scaling(x):
        return np.log1p(np.log1p(np.log1p(x)))

    def mate_internal(self, item1: Item, item2: Item) -> Genes:
        '''Intersection mating with weighted information gain.

        First Calculates intersection of genes. Then featuresare sorted
        by information gain and best (how many is decided by number
        function) features by information gain are added to the set
        with a probability proportionate to the information gain.

        Args:
            item1: An Item object we wish to mate.
            item2: An Item object we wish to mate.

        Returns:
            An array of booleans (genes) that is the result of mating.
        '''
        genes = super().mate_internal(item1, item2)
        ma = max(item1.size, item2.size)
        mi = min(item1.size, item2.size)

        sy_difference = Item. \
            indexed_symmetric_difference(item1.genes, item2.genes)

        weights = \
            self.scaling([self.scikit_information_gain[i] for i in sy_difference])

        if sy_difference:  # If there are any different genes
            # They might all have 0 information gain
            if weights.sum() == 0:
                weights[:] = np.ones_like(weights) / len(weights)
            for ind in random_utils.choices(sy_difference, p=weights / weights.sum(),
                                            size=self.number(ma - mi)):
                genes[ind] = True

        return genes


@dataclass
class MatingPoolResult:
    '''Mating pool result.

    A class used to store mating pool and items to be promoted to the
    next generation.

    Attributes:
        mating_pool: A list of EvalItem objects that are to be mated.
        carry_over: A list of EvalItem objects from current generation
            to be promoted to the next one.
    '''
    mating_pool: List["EvalItem"]
    carry_over: List["EvalItem"]


class MatingSelectionStrategy(ABC):
    '''Mating selection strategy.

    An abstract class that defines how the current population will be
    processed (in terms of mating) but not how the mating pool will be
    choosen or how pairs from that pool will be generated.

    Attributes:
        mating_strategy: Mating strategy to be used in the algorithm.
    '''
    def __init__(self, mating_strategy: MatingStrategy,
                 overwrite_carry_over: bool = True
                 ) -> None:
        self.mating_strategy = mating_strategy
        self.overwrite_carry_over = overwrite_carry_over

    def process_population(self, population: Population,
                           current_generation_number: int) -> \
            UnevaluatedPopulation:
        '''Procesess mating of population.

        Choses the mating pol and generates mated population then joins
        the items from the last generation (that will be in the new
        generation ass well) with the items that are the result of
        mating.

        Args:
            population: A dictionary that represents the Current
                population (of EvalItem objects).
            current_generation_number: An integer which represents the
                current generation of mating.

        Returns:
            A dictionary that represents the new population
            (of Item objects).
        '''
        result = self.mating_pool(population)
        mated = self.mate_pool(result.mating_pool, current_generation_number)

        new_population: UnevaluatedPopulation = {}

        for last in result.carry_over:
            new_population.setdefault(last.size, []).append(last)
        for new in mated:
            new_population.setdefault(new.size, []).append(new)

        return new_population

    @abstractmethod
    def mating_pool(self, population: Population) -> MatingPoolResult:
        pass

    @abstractmethod
    def mate_pool(self, mating_pool: List["EvalItem"],
                  current_generation: int = 0) -> List[Item]:
        pass

    def use_data_information(self, train_data: np.array,
                             train_target: np.array) -> None:
        self.mating_strategy.use_data_information(train_data, train_target)


class RandomEveryoneWithEveryone(MatingSelectionStrategy):
    '''Everyone with everyone mating selection strategy.

    A subclass of MatingSelectionStrategy in which a mating pool of
    size pool_size is chosen and every possible pair from this pool is
    mated.

    Attributes:
        pool_size: the size of the pool from which mating pairs are
            generated.
    '''
    def __init__(self, mating_strategy: MatingStrategy = UnionMating(),
                 pool_size: int = 10):
        super().__init__(mating_strategy)
        self.pool_size = pool_size

    def mating_pool(self, population: Population) -> MatingPoolResult:
        '''Chooses mating pool.

        Transforms population into a single list and randomly chooses
        a mating pool of size pool_size.

        Args:
            population: A dictionary with List[Item] values and form
                those Items the mating pool is choosen.

        Returns:
            A MatingPoolResult object.
        '''
        flatten = flatten_population(population)
        pool_size = self.pool_size if self.pool_size is not None else len(
            flatten)
        return MatingPoolResult(random_utils.choices(flatten, size=self.pool_size),
                                flatten)

    def mate_pool(self, mating_pool: List["EvalItem"],
                  current_generation: int = 0) -> List[Item]:
        '''Chooses each possible pair from mating pool and mates them.

        Iterates through all possible pairs of items in mating pool and
        calls mate (from choosen mating strategy) on the pair.
        Generation of the offspring is increased by 1.

        Args:
            mating_pool: A List of EvalItem objects from which pairs
                for mating will be chosen.
            current_generation: An integer which represents the current
                generation of mating.

        Returns:
            A List of Item objects (the result of mating).
        '''
        rtr: List["Item"] = []
        for i in range(len(mating_pool)):
            for j in range(i + 1, len(mating_pool)):
                rtr.append(
                    self.mating_strategy.mate(mating_pool[i], mating_pool[j],
                                              current_generation + 1))
        return rtr


class NoMating(MatingSelectionStrategy):
    '''No mating selection strategy.

    A subclass of MatingSelectionStrategy which defines a strategy
    where no mating is performed.
    '''

    def mating_pool(self, population: Population) -> MatingPoolResult:
        '''Creates a MatingPoolResult object.

        Accepts a population (a dictionary) and makes that population
        a mating pool.

        Args:
            population: A dictionary to be made a mating pool.

        Returns:
            A MatingPoolResult object with population as mating pool.
        '''
        f_pop = flatten_population(population)
        #assert len(f_pop) == len(population)
        return MatingPoolResult(
            f_pop, []
        )

    def mate_pool(self, mating_pool: List["EvalItem"],
                  current_generation: int = 0) -> List[Item]:
        '''Raises the generation number of items without mating them.

        Iterates through EvalItem objects in the list and creates new
        Item objects with the generation number increased by one and
        appends them to an array.

        Argy:
            mating_pool: A list of EvalItem objects.
            current_generation: An integer which represents the current
                generation of mating.

        Returns:
            A List of Item objects (equivalent to the input list but
            with increased generation).
        '''
        return [Item(item.genes, item.generation + 1, None, None, ) for item in
                mating_pool]

    def use_data_information(self, train_data: np.array,
                             train_target: np.array) -> None:
        '''Overrides method from superclass.
        '''
        pass


class MutationStrategy(ABC):
    '''Mutation strategy

    An abstract class that defines how the population will be processed
    (mutated) and how mutation of each item will be performed but not
    the actual mutation implementation.
    '''
    @abstractmethod
    def mutate_internal(self, item: "Item") -> Genes:
        pass

    def process_population(self, population: UnevaluatedPopulation) -> \
            UnevaluatedPopulation:
        '''Iterates over population and performs mutations.

        Iterates over dictionary entries and for each item
        (list of Item objects) again iterates through Item objects in
        it, performs mutation and appends it to the correct List in
        the dictionary (or creates a new one if it does not exist).

        Args:
            population: A dictionary with integers (Item.size) as keys and lists
                of items of that size.

        Returns:
            A dictionary with the same content as the input dictionary
            but with mutated Items.
        '''
        pop2: UnevaluatedPopulation = {}
        for num, items in population.items():
            for item in items:
                mutated = self.mutate(item)
                pop2.setdefault(mutated.size, []).append(mutated)

        return pop2

    def mutate(self, item: "Item") -> Item:
        '''Mutates a single item.

        Performs mutation on an item (actual mutation is to be
        implemented) and from creates a new Item object form the
        mutates genes.

        Args:
            item: The item to be mutated.

        Returns:
            An Item object with mutated genes.
        '''
        new_genes = self.mutate_internal(item)

        return Item(new_genes, item.generation, item.parent_a, item.parent_b)


class RandomFlipMutationStrategy(MutationStrategy):
    '''Random flip mutation strategy.

    A subclass of MutationStrategy class that defines mutation as a
    flip of a gene with some probability.

    Attributes:
        prob: A float representing the probability of a gene beeing
            flipped
    '''
    def __init__(self, prob: float = 1 / 100):
        self.prob = prob

    def mutate_internal(self, item: "Item") -> Genes:
        '''Mutation implementation.

        Loops over given genes and flips each one with probability
        prob.

        Args:
            item: An item object whose genes we want to mutate.

        Returns:
            An array of booleans (mutated genes).
        '''
        genes = [
            not j if random_utils.random() < self.prob else j
            for j in item.genes
        ]
        return genes



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='fastener',
    version='1.0.3',
    author='Filip Koprivec, Klemen Kenda, Gal Petkovsek',
    author_email='',
    description='Feature selection enabled by entropy',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JozefStefanInstitute/FASTENER.git",
    packages=["fastener"],
    #py_modules=["fastener", "item", "random_utils"],
    install_requires=["joblib==0.14.1",
                    "mypy==0.761",
                    "mypy-extensions==0.4.3",
                    "numpy==1.18.1",
                    "scikit-learn>=0.22.2",
                    "scipy>=1.4.1",
                    "typed-ast>=1.4.1",
                    "typing-extensions>=3.7.4.1"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
    ],
    license='MIT',
    python_requires='>=3.6',
    package_dir = {'': 'src'},
)


'''
FASTENER main classes.
'''

import math
import os
import pickle
import time
from dataclasses import dataclass, field
from functools import wraps

import fastener.random_utils as random_utils
from fastener.random_utils import shuffle
from typing import Dict, List, Callable, Any, Tuple, Optional, \
    Counter as CounterType, Set
import numpy as np
from collections import Counter

from fastener.item import Item, Result, Population, flatten_population, FitnessFunction, \
    Genes, EvalItem, RandomFlipMutationStrategy, RandomEveryoneWithEveryone, \
    IntersectionMating, UnionMating, IntersectionMatingWithInformationGain, \
    IntersectionMatingWithWeightedRandomInformationGain, UnevaluatedPopulation, \
    MatingStrategy, MutationStrategy, MatingSelectionStrategy

Front = Dict[int, EvalItem]


@dataclass
class LogData:
    '''Represents a object to be saved into log

    Contains data about each generation that will be saved into log.

    Attributes:
        generation: An integer representing the generation the log is
            about.
        population: A Population object which contains population of
            current generation.
        front: A dictionary holding the best Item (from the
            population) object for each number of selected features.
        mated: Population after mating.
        mutated: Population after mutation.
        cache_counter: A dictionary, counting how many times each value
            was cought.
        cache_data: A dictionary that saves score for every set of
            genes calculated.
        random_state: A random state to be saved in the log file.
        timestamp_step: A float representing the time when the object
            was created.

    '''
    generation: int
    population: Population
    front: Front
    mated: UnevaluatedPopulation
    mutated: UnevaluatedPopulation

    cache_counter: CounterType[int]
    cache_data: Dict[int, Result]
    random_state: Any
    timestamp: float = field(init=False)

    def __post_init__(self):
        self.timestamp = time.time()

    @staticmethod
    def discard_model(cache_data: Dict[int, Tuple[Result, Any]]) -> Dict[int, Result]:
        return {num: item[0] for num, item in cache_data.items()}

    def dump_log(self, config: "Config") -> None:
        '''Saves self into a pickle file.

        Saves the instance of the current object into output folder
        as a pickle file.

        Args:
            config: A Config object in which the output folder name is
                stored.
        '''
        pickle.dump(
            self,
            open(os.path.join(
                config.output_folder,
                f"generation_{self.generation}.pickle"),
                "wb")
        )


@dataclass
class Config:
    '''Configuration parameters

    An object that stores configuration parameters of the algorithm.

    Attributes:
        output_folder: A string containing the name of the output
            folder for log.
        rendom_seed: Used to initialize seed from random_utils.
        number_of_rounds: An integer representing the number of rounds
            (generations) of the algorithm.
        max_bucket_size: An  integer representing the maximum number
            of items with the same number of features.
        reset_to_pareto_rounds: An integer which represents the
            interval when the population is set to front.
        cache_fitness_function: A boolean determining if evaluation
            results will be cached
    '''
    output_folder: str
    random_seed: int
    number_of_rounds: int = 1000  # Number of round of genetic algorithm
    max_bucket_size: int = 3  # Max number of items in same size bucket
    reset_to_pareto_rounds: Optional[int] = 5  # Reset population to Pareto front every n-rounds (False is never)

    cache_fitness_function: bool = True

    def __post_init__(self) -> None:
        '''Sets atributes.

        If reset_to_pareto_rounds is not defined it is set to
        number_of_rounds. Output folder name is set and random seed is
        asigned.
        '''
        if not self.reset_to_pareto_rounds:
            self.reset_to_pareto_rounds = self.number_of_rounds
        self.output_folder = os.path.join("log", self.output_folder)

        random_utils.seed(self.random_seed)


class EntropyOptimizer:
    '''Entropy Optimizer.

    This class holds variables that describe the settings of the
    algorithm and also holds information about current generation that
    is beeing used in the algorithm.

    Attributes:
        model: A model used for evaluation of the selected features.
        train_data: The data the model is training on.
        train_target: The target values for training the model.
        evaluator: A function that accepts the "genes", tests that set
            of genes and returns Result.
        number_of_genes: An integer representing the number of genes.
        mating_selection_strategy: A MatingSelectionStrategy object
            that defines the strategy used for mating.
        reset_to_front_predicate: TODO
        initial_population: An optional attribute containing the
            population the algorithm starts with.
        initial_genes: An optional attribute containing a 2D array of
            integers with indices of active genes for constructing
            initial population.
        config: A Config class with settings of the algorithm.
        cache_data: A dictionary that saves score for every set of
            genes calculated.
        cache_counter: A dictionary, counting how many times each value
            was cought.
        population: The population of the current state of algorithm.
        pareto_front: A dictionary holding the best Item (from the
            population) object for each number of selected features.
        fitness_function: A function that returns the score for a given
            gene.
    '''
    def __init__(self,
                 model: Any,
                 train_data: np.array,
                 train_target: np.array,
                 evaluator: Callable[
                     [Any, "Genes", Optional[List[int]]], "Result"],
                 number_of_genes: int,
                 mating_selection_strategy: MatingSelectionStrategy,
                 mutation_strategy: MutationStrategy,
                 reset_to_front_predicate: Optional[
                     Callable[[int, Population, Front], bool]] = None,
                 initial_population: Optional[UnevaluatedPopulation] = None,
                 initial_genes: Optional[List[List[int]]] = None,
                 config: Optional[Config] = None) -> None:
        self._model = model
        self.evaluator = evaluator

        self.train_data = train_data
        self.train_target = train_target

        self.number_of_genes = number_of_genes
        if config is None:
            config = Config("data", 2020)

        self.config = config
        os.makedirs(self.config.output_folder, exist_ok=False)

        self.mating_selection_strategy = mating_selection_strategy
        self.mutation_strategy = mutation_strategy

        self.initial_population = initial_population
        self.initial_genes = initial_genes

        if reset_to_front_predicate is None:
            self.reset_to_front_predicate = EntropyOptimizer.default_reset_to_pareto
        else:
            self.reset_to_front_predicate = reset_to_front_predicate

        self.cache_counter: CounterType[int] = Counter()
        self.cache_data: Dict[int, Tuple[Result, Any]] = {}

        self.population: Population = {}
        self.pareto_front: Front = {}

        self.fitness_function = self._fitness_function

    def train_model(self, genes: "Genes") -> Any:
        '''Trains the model.

        Trains the model on the training data (using only features
        definded by genes array and train target).

        Args:
            genes: An array of booleans (genes) that detemines which
                features will be considered when training the  model

        Returns:
            The trained model.
        '''
        return self._model().fit(self.train_data[:, genes], self.train_target)

    def _fitness_function(self, genes: "Genes") -> "Tuple[Result, Any]":
        '''evaluates trained model.

        Calls function to train the model and then evaluates the model
        with the specified evaluation function.

        Args:
            genes: An array of booleans (genes) that represent the
                model to be evaluated.

        Returns:
            A tuple of result (from evaluation) and the trained model
        '''
        model = self.train_model(genes)
        return self.evaluator(model, genes, None), model

    def cached_fitness(self, genes: "Genes") -> "Tuple[Result, Any]":
        '''Same as _fitnes_function but uses cached values if it can.

        For the given set of genes it checks if cache_data contains
        that record already. If it does not it creates a new tuple
        with _fitness_function and saves it to cache_data.

        Args:
            genes: An array of booleans (genes) whose cache_data record
                is to be returned

        Returns:
            A tuple[Result, Any] that was obtained from cache_data or
            calculated.
        '''
        #print("Trying:", np.where(genes))
        number = Item.to_number(genes)
        result = self.cache_data.get(number, None)
        self.cache_counter[number] += 1
        if result is None:
            result = self._fitness_function(genes)
            self.cache_data[number] = result
        else:
            #print("Hitted", self.cache_counter)
            pass
        return result

    def purge_front_with_information_gain(self):
        '''Purges front.

        To each item in the front it removes the worst gene and adds
        that item to the front if it is better than the one already in
        the front or there is no item with that number of features in
        the front.

        Args:

        Returns:

        Raises:
            AssertionError: If the number of features in the new item
                is not one less than before.
        '''
        new_items = []
        for num, item in self.pareto_front.items():
            if num == 1:  # Can't remove features
                continue
            new_item = self.purge_item_with_information_gain(item)
            assert new_item.size == num - 1
            new_items.append(new_item)

        better = 0
        new = 0
        for item in new_items:
            if item.size in self.pareto_front:
                if item.result > self.pareto_front[item.size].result:
                    self.pareto_front[item.size] = item
                    better += 1
            else:
                self.pareto_front[item.size] = item
                new += 1

        # print("Purged:", "better:", better, "new:", new)

        self.remove_pareto_non_optimal()

    def purge_item_with_information_gain(self, item: "Item") -> "EvalItem":
        '''Removes the worst gene from an item.

        Checks which gene adds the least to the score and removes it
        from the item.

        Args:
            item: An Item object to which we want to remove the worst
                gene.

        Returns:
            An EvalItem Object that is the same as the input object but
            with the worst gene removed.
        '''
        base_result, model = self.fitness_function(item.genes)
        on_genes = np.where(item.genes)[0]
        changes = [(1.0, -1) for _ in on_genes]

        for change_ind, gene_ind in enumerate(on_genes):
            sh_result = self.evaluator(model, item.genes, [change_ind])
            changes[change_ind] = (base_result.score - sh_result.score, gene_ind)
        changes.sort()  # Sort them so that first cause the smallest change
        # print(base_result)
        # print("CHANGES:", changes)
        # And therefore seem a good fit for removal
        # Maybe some random selection?
        genes2 = item.genes.copy()
        # Unset the gene with the smallest change
        genes2[changes[0][1]] = False

        rt_item = Item(genes2, item.generation + 1, None, None).\
            evaluate(self.fitness_function)
        return rt_item

    @staticmethod
    def default_reset_to_pareto(round_num, _population, _front):
        return round_num % 3 == 0

    def prepare_loop(self) -> None:
        '''Preparations befor running main loop.

        It sets some parameters and initial population if it is given
        and also creates new EvalItem objects for given initial genes.
        It also saves current EntropyOptimizer object in a pickel file.

        Args:

        Returns:

        Raises:
            AssertionError: If initial_population and initial_genes are
                not provided.
        '''
        if self.config.cache_fitness_function:
            self.fitness_function = self.cached_fitness

        if self.initial_population:
            self.population = self.evaluate_unevaluated(self.initial_population)

        if self.initial_genes:
            for in_gene in self.initial_genes:
                zeros = np.zeros(self.number_of_genes, dtype=bool)
                zeros[in_gene] = 1
                itm = Item(list(zeros), 0, None, None).evaluate(self.fitness_function)
                self.population.setdefault(itm.size, []).append(itm)

        assert self.population, "Some sort of initial population must be provided"

        self.mating_selection_strategy.use_data_information(self.train_data,
                                                            self.train_target)

        pickle.dump(
            self, open(
                os.path.join(self.config.output_folder, "experiment.pickle"),
                "wb")
        )

    def mainloop(self) -> None:
        '''Main loop of the algorithm.

        Consists of preparation part and the loop which loops as many
        times as it is specified by the total number of generations.
        In each loop occures mating then mutation and after that
        removal of duplicates, evaluation of the new population. The
        front is also updated and then excessive and not optimal items
        are removed from it. If required (depending on
        reset_to_pareto_rounds and round_n) the
        purge_front_with_information_gain function is executed and the
        population is set to front. At the end of each loop the current
        EntropyOptimizer object is saved in a pickle file.

        Args:

        Returns:
        '''

        self.prepare_loop()

        self.purge_oversize_buckets()
        self.update_front_from_population()

        for round_n in range(1, self.config.number_of_rounds + 1):
            print(f"Round: {round_n}")
            mated = self.mating_selection_strategy.process_population(
                self.population, round_n
            )
            mutated = self.mutation_strategy.process_population(mated)
            mutated = self.clear_duplicates_after_mating(mutated)
            # print(mutated)
            # print(len(mutated))
            # Evaluate new population
            # Do not evaluate "empty" genes
            self.population = self.evaluate_unevaluated(mutated)

            self.purge_oversize_buckets()
            self.update_front_from_population()
            bef = len(self.pareto_front)
            self.remove_pareto_non_optimal()
            aft = len(self.pareto_front)

            # print(len(flatten_population(self.population)), bef, aft)


            if self.config.reset_to_pareto_rounds and \
                    round_n % self.config.reset_to_pareto_rounds == 0:

                self.purge_front_with_information_gain()

                self.reset_population_to_front()
                # print("RESETING POPULATION TO FRONT")

            log_data = LogData(round_n, self.population, self.pareto_front, mated,
                               mutated, self.cache_counter,
                               LogData.discard_model(self.cache_data),
                               random_utils.get_state())

            # Log during evaluation
            log_data.dump_log(self.config)

    def purge_oversize_buckets(self) -> None:
        '''Removes excessive Items.

        For each number of features it filters the items and keeps only
        the ones with the best score (how many is decided with
        max_bucket_size).

        Args:

        Returns:
        '''
        new_population: Population = {}
        for num, items in self.population.items():
            new_population[num] = sorted(items, reverse=True)[:self.config.max_bucket_size]
        self.population = new_population

    def update_front_from_population(self) -> None:
        '''Updates front.

        Assuming that the population is sorted (the first elements in
        lists score the highest), it loops through population and for
        each number of features checks if item in pareto_front[num]
        scores higher (or the same) than the best one in population and
        if not it updates it. If no element with that number of
        features is yet in pareto_front it is also added.

        Args:

        Returns:
        '''
        # Assume that population is sorted
        for num, items in self.population.items():
            if num in self.pareto_front:
                if items:
                    self.pareto_front[num] = \
                        max(self.pareto_front[num], items[0])
            else:
                if items:
                    self.pareto_front[num] = items[0]

    @classmethod
    def clear_duplicates_after_mating(cls, un_pop: "UnevaluatedPopulation") \
            -> "UnevaluatedPopulation":
        '''Removes duplicates from population.

        Creates a new unevaluated population and adds in only one
        instance of each Item in unevaluated population.

        Args:
            un_pop: Unevaluated population that needs to be cleared of
                duplicates.

        Returns:
            Unevaluated population, cleared of duplicates.
        '''
        new_un_pop: UnevaluatedPopulation = {}
        for num, items in un_pop.items():
            new: List[Item] = []
            taken: Set[int] = set()
            for item in items:
                if item.number not in taken:
                    taken.add(item.number)
                    new.append(item)
            new_un_pop[num] = new

        return new_un_pop

    def remove_pareto_non_optimal(self) -> None:
        '''Removes not optimal mambers of the front

        Removes items from the front that have worse or equal score as
        than an item with less features.

        Args:

        Returns:
        '''
        # Make a copy as to not change dict during looping,
        # sort them by the number of genes
        sorted_items = list(sorted(
            self.pareto_front.items(), key=lambda item: item[0]
        ))

        current_peak = sorted_items[0][1]
        for j in range(1, len(sorted_items)):
            n, item = sorted_items[j]
            if current_peak.pareto_better(item):
                del self.pareto_front[n]
            else:
                current_peak = item

    def reset_population_to_front(self) -> None:
        '''Sets population to front.

        The current front is copied into population.

        Args:

        Returns:
        '''
        new_population: Population = {}
        # Keep Pareto front in itemized order
        for num, item in sorted(self.pareto_front.items(),
                                key=lambda itm: itm[0]):
            new_population[num] = [item]
        self.population = new_population

    def evaluate_unevaluated(self, un_pop: "UnevaluatedPopulation") \
            -> Population:
        '''Evaluates population.

        Takes dictionary of unevaluated items and creates dictionary
        of evaluated items.

        Args:
            un_pop: A dictionary of unevaluated population.

        Returns:
            A dictionary of evaluated population.
        '''
        return {
            num: [item.evaluate(self.fitness_function) for item in items
                  if item.size]
            for num, items in un_pop.items()
        }


"""from sklearn.tree import DecisionTreeClassifier
from sklearn.metrics import f1_score

general_model = DecisionTreeClassifier
from src.dummy_data import XX_train, XX_test, labels_train, labels_test


def eval_fun(model: Any, genes: "Genes",
             shuffle_indices: Optional[List[int]] = None) -> "Result":
    test_data = XX_test[:, genes]
    if shuffle_indices:
        test_data = test_data.copy()
        for j in shuffle_indices:
            shuffle(test_data[:, j])
    pred = model.predict(test_data)
    res = Result(f1_score(labels_test, pred, average='weighted'))
    return res


def main() -> None:

    number_of_genes = XX_train.shape[1]

    initial_genes = [
        [0]
    ]

    # Select mating strategies
    mating = RandomEveryoneWithEveryone(
        pool_size=3,
        mating_strategy=IntersectionMatingWithWeightedRandomInformationGain())

    # Random mutation
    mutation = RandomFlipMutationStrategy(1 / number_of_genes)

    entropy_optimizer = EntropyOptimizer(
        general_model, XX_train, labels_train, eval_fun,
        number_of_genes, mating, mutation, initial_genes=initial_genes,
        config=Config(output_folder="dummy_data_output", random_seed=2020,
                      reset_to_pareto_rounds=5)
    )

    entropy_optimizer.mainloop()


if __name__ == '__main__':
    main()
"""